<?php

use Controller\ErrorController;
use Shopbydedi\ErrorMiddleWare;
use Shopbydedi\StackMiddleware;
use Symfony\Component\Debug\ErrorHandler;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\EventDispatcher\EventDispatcher;
use Symfony\Component\Debug\ExceptionHandler;
use Symfony\Component\Routing\Exception\ResourceNotFoundException;

use Auth\Auth;
use Auth\AuthException;
use Event\ApplicationEvent;
use Detection\MobileDetect;

/**
 * Class Application
 */
class Application
{

    protected static $booted = false;

    /**
     * @var \DebugBar\StandardDebugBar()
     */
    protected static $debugBar;

    /**
     * @var Router
     */
    protected static $router;

    /**
     * @var \Symfony\Component\EventDispatcher\EventDispatcher
     */
    protected static $eventDispatcher;

    /**
     * @var \Symfony\Component\HttpFoundation\Request
     */
    protected static $request;

    /**
     * @var \Auth\Auth
     */
    protected static $auth;

    /**
     * @var EmailSender[]
     */
    protected static $emailSender;

    /**
     * @var \Detection\MobileDetect
     */
    protected static $mobileDetect;

    /**
     * Get the debug bar
     *
     * @return \DebugBar\DediDebugBar
     */
    public static function getDebugBar()
    {
        return self::$debugBar;
    }

    /**
     * @return Router
     */
    public static function getRouter()
    {
        return self::$router;
    }

    /**
     * @return \Symfony\Component\EventDispatcher\EventDispatcher
     */
    public static function getEventDispatcher()
    {
        return self::$eventDispatcher;
    }

    /**
     * @return Request
     */
    public static function getCurrentRequest()
    {
        return self::$request;
    }

    /**
     * @return Auth
     */
    public static function getAuth()
    {
        return self::$auth;
    }

    /**
     * @return \Detection\MobileDetect
     */
    public static function getMobileDetect()
    {
        return self::$mobileDetect;
    }

    /**
     * @param $name
     *
     * @return EmailSender
     */
    public static function getEmailSender($name = 'default')
    {
        return self::$emailSender[$name];
    }

    /**
     * Launch the app
     * @throws DediException
     */
    protected static function boot()
    {

        if (getenv('APP_DEBUG')) {
            ini_set('display_errors', 1);
            ini_set('error_reporting', E_ALL);

            // check si on est pas en ligne de commande (php-cli)
            if ('cli' !== php_sapi_name()) {
                // permet de catcher les exceptions qui ne le sont pas
                // et affichage une page d'erreur
                ExceptionHandler::register();
                ErrorHandler::register();
            }

            // permet de transformer les erreur PHP en Exception
            // ErrorHandler::register();
        } else {
            ini_set('display_errors', 0);
            ini_set('error_reporting', E_ALL ^ E_NOTICE);
        }

        self::$eventDispatcher = new EventDispatcher();

        if (getenv('APP_DEBUG') && class_exists('DebugBar\\DediDebugBar')) {
            // instanciate the debug bar
            self::$debugBar = new \DebugBar\DediDebugBar();
        }

        // register events
        include __DIR__ . '/../../config/events.php';

        /**
         * Code après peut utiliser les events, car enregistré plus haut
         */

        self::$router = new Router();

        if (getenv("APP_PREFIX") !== false) {
            self::$router->getRouter()->getRouteCollection()->addPrefix(getenv("APP_PREFIX"));
        }

        // load Module
        (new \Module\ModuleLoader(self::getRouter(), self::getEventDispatcher()))
            ->setModule(include __DIR__ . '/../../config/modules.php')
            ->load();

        // load Alias
        (new \Alias\AliasLoader())
            ->setAlias(include __DIR__ . '/../../config/alias.php')
            ->load();

        self::$emailSender = array(
            'default' => new EmailSender(self::getTransport())
        );

        self::$mobileDetect = new MobileDetect();

        self::$booted = true;
    }

    /**
     * @return Swift_Transport
     */
    protected static function getTransport()
    {
        switch (getenv('EMAIL_TRANSPORT')) {
            case 'smtp':
                $transport = Swift_SmtpTransport::newInstance(getenv('EMAIL_HOST'), getenv('EMAIL_PORT'));
                $transport->setUsername(getenv('EMAIL_USER'));
                $transport->setPassword(getenv('EMAIL_PASSWORD'));
                break;
            default:
                $transport = Swift_MailTransport::newInstance();
                break;
        }

        return $transport;
    }

    /**
     * boot the app if not booted yet
     * @throws DediException
     */
    public static function bootIfNotBooted()
    {
        if (self::$booted == false) {
            self::boot();
        }
    }

    /**
     * Handle a request and return a response
     *
     * @param Request $request the request object
     *
     * @throws DediException
     * @throws SmartyException
     *
     * @return Response
     */
    public static function handle(Request $request)
    {
        self::bootIfNotBooted();

        self::$auth = Auth::getInstance();

        // Langue::handle();
        Devise::handle();

        $boutique = Boutique::getInstance();

        $langue = $boutique->getLangue();

        // init constant
        define('SITE_NAME', $boutique->nom);
        define('ADMIN_EMAIL', $boutique->adminEmail);
        define('ADMIN_NAME', $boutique->nom);
        define('CONTACT_EMAIL', $boutique->contactEmail);
        define('CONTACT_EMAIL_NAME', $boutique->nom);
        define('META_LANG', strtolower($langue->ini));

        // ajout pour format date en fonction de la langue du pays
        date_default_timezone_set('Europe/Paris');
        setlocale(LC_TIME, $langue->ini . '_' . strtoupper($langue->ini) . '.utf8');

        // on track le referer
        Tracking::initReferer();

        // on met a jour le context en fonction de la requete
        self::$router->getRouter()->getContext()->fromRequest($request);

        $middleWares = [];
        if (!getenv('APP_DEBUG')) {
            $middleWares[] = new ErrorMiddleWare(Tools::isDediServicesIp(), new ErrorController($request));
        }

        $middleWare = new StackMiddleware(...$middleWares);

        $response = $middleWare->process($request, function (Request $request) {
            $response = Redirection::handle($request);
            if ($response instanceof RedirectResponse) {
                return $response;
            }

            try {
                // on match la request avec une url
                $request->attributes->add(self::$router->getRouter()->matchRequest($request));

                self::$request = $request;

                self::$auth->setRequest($request);

                $role = $request->attributes->get('role');
                if (!empty($role)) {
                    $response = self::$auth->checkRole($role);
                    if ($response instanceof Response) {
                        return $response;
                    }
                }

                $controller = $request->attributes->get('_controller');

                self::$eventDispatcher->dispatch(
                    ApplicationEvent::BEFORE_CALL_CONTROLLER,
                    new ApplicationEvent($request, new Response())
                );

                $response = call_user_func_array([new $controller($request), $request->attributes->get('_action')], []);

            } catch (ResourceNotFoundException $e) {
                // si on est en mode debug, on relance l'exception pour que le gestionnaire la traite
                if (getenv('APP_DEBUG')) {
                    throw $e;
                } else {

                    if (empty(self::$request)) {
                        self::$request = $request;
                    }

                    // si en prod, on affiche une page 404
                    $response = (new ErrorController($request))->error_404();
                }
            } catch (AuthException $e) {
                $response = $e->getResponse();
            }

            if (empty($response) || !$response instanceof Response) {
                if (getenv('APP_DEBUG')) {
                    throw new DediException(
                        sprintf(
                            'no response return in %s::%s()',
                            $request->attributes->get('_controller'),
                            $request->attributes->get('_action')
                        )
                    );
                } else {
                    $response = new Response();
                }
            }

            return $response;
        });

        self::$eventDispatcher->dispatch(
            ApplicationEvent::BEFORE_REPSONSE,
            new ApplicationEvent($request, $response)
        );

        return $response;
    }

    /**
     * Send event terminate with request and response
     *
     * @param Request  $request  the current request
     * @param Response $response the reponse
     *
     * @return void
     */
    public static function terminate(Request $request, Response $response)
    {
        self::$eventDispatcher->dispatch(
            ApplicationEvent::TERMINATE,
            new ApplicationEvent($request, $response)
        );
    }

    /**
     * Render the debug bar
     *
     * @return void
     */
    public static function renderDebugBar()
    {
        $render = self::$debugBar->getJavascriptRenderer();

        echo $render->renderHead();
        echo $render->render();
    }
}
